#include <intrin.h>
#include "xts.h"

#define Mult(T) \
{	__m128i tmp = _mm_shuffle_epi32(_mm_srli_epi64(T, 63), _MM_SHUFFLE(1, 0, 3, 2)); \
	T = (_mm_slli_epi64(T, 1) | tmp) ^ _mm_mul_epu32(tmp, GF_128_FDBK); }

void XTS_EncryptSector(int cipher, unsigned long long sector, void *buffer)
{
	void *key1, *key2;
	__m128i T[4], *ptr;
	unsigned int i, ki = n_cipher;
	__m128i GF_128_FDBK = _mm_set_epi64x(0, 0x87 - 1);

	do {
		ptr = buffer;
		ki -= 2;
		key1 = &keys[ki];
		key2 = &keys[ki + 1];

		T[0] = _mm_movpi64_epi64((__m64)sector);

		if ((unsigned char)cipher == CIPHER_AES) {
			aes_encrypt(key2, T);
			for (i = 0; i < 512 / 16; i++, ptr++) {
				*ptr ^= T[0];
				aes_encrypt(key1, ptr);
				*ptr ^= T[0];
				Mult(T[0]);
			}
		}
		else if ((unsigned char)cipher == CIPHER_SERPENT) {
			serpent_encrypt(key2, T);
			for (i = 0; i < 512 / 64; i++, ptr += 4) {
				T[1] = T[0];
				Mult(T[1]);
				T[2] = T[1];
				Mult(T[2]);
				T[3] = T[2];
				Mult(T[3]);
				*ptr ^= T[0];
				*(ptr + 1) ^= T[1];
				*(ptr + 2) ^= T[2];
				*(ptr + 3) ^= T[3];
				serpent_encrypt(key1, ptr);
				*ptr ^= T[0];
				*(ptr + 1) ^= T[1];
				*(ptr + 2) ^= T[2];
				*(ptr + 3) ^= T[3];
				T[0] = T[3];
				Mult(T[0]);
			}
		}
		else {
			twofish_encrypt(key2, T);
			for (i = 0; i < 512 / 16; i++, ptr++) {
				*ptr ^= T[0];
				twofish_encrypt(key1, ptr);
				*ptr ^= T[0];
				Mult(T[0]);
			}
		}
	} while (cipher >>= 8);
}

void XTS_DecryptSector(int cipher, unsigned long long sector, void *buffer)
{
	void *key1, *key2;
	__m128i T[4], *ptr;
	unsigned int i, ki = 0;
	__m128i GF_128_FDBK = _mm_set_epi64x(0, 0x87 - 1);

	do {
		ptr = buffer;
		key1 = &keys[ki];
		key2 = &keys[ki + 1];

		T[0] = _mm_movpi64_epi64((__m64)sector);

		if ((unsigned char)cipher == CIPHER_AES) {
			aes_encrypt(key2, T);
			for (i = 0; i < 512 / 16; i++, ptr++) {
				*ptr ^= T[0];
				aes_decrypt(key1, ptr);
				*ptr ^= T[0];
				Mult(T[0]);
			}
		}
		else if ((unsigned char)cipher == CIPHER_SERPENT) {
			serpent_encrypt(key2, T);
			for (i = 0; i < 512 / 64; i++, ptr += 4) {
				T[1] = T[0];
				Mult(T[1]);
				T[2] = T[1];
				Mult(T[2]);
				T[3] = T[2];
				Mult(T[3]);
				*ptr ^= T[0];
				*(ptr + 1) ^= T[1];
				*(ptr + 2) ^= T[2];
				*(ptr + 3) ^= T[3];
				serpent_decrypt(key1, ptr);
				*ptr ^= T[0];
				*(ptr + 1) ^= T[1];
				*(ptr + 2) ^= T[2];
				*(ptr + 3) ^= T[3];
				T[0] = T[3];
				Mult(T[0]);
			}
		}
		else {
			twofish_encrypt(key2, T);
			for (i = 0; i < 512 / 16; i++, ptr++) {
				*ptr ^= T[0];
				twofish_decrypt(key1, ptr);
				*ptr ^= T[0];
				Mult(T[0]);
			}
		}

		ki += 2;
	} while (cipher >>= 8);
}
